/**********************************************************************\
*                                                                      *
*                  Copyright (c) Exeter University,		       *
*                  Department of Computer Science,                     *
*                 (aba@uk.ac.ex.dcs) Adam Back, 1994                   *
*                                                                      *
*  GNU GPL (General Public License) Applies.                           *
*                                                                      *
\**********************************************************************/
/*
 * Transputer XDR
 *
 * Defines the following types:
 *
 *	XDR;
 *	bool_t;
 *	u_int;
 *      u_long;
 *      u_short;
 *	caddr_t;
 *	xdrproc_t
 *	enum xdr_op;
 *
 * And the following constants:
 *
 *      XDR_ENCODE
 *      XDR_DECODE
 *      XDR_FREE
 *
 * And the following functions/macros:
 *
 *  SIMPLE:
 *
 *	bool_t	xdr_short(XDR *xdrs, short *sp);
 *	bool_t	xdr_int(XDR *xdrs, int *ip);
 *	bool_t	xdr_long(XDR *xdrs, long *lp);
 *	bool_t	xdr_array(XDR *xdrs, caddr_t *arrp, u_int *sizep,
 *			  u_int maxsize, u_int elsize, xdrproc_t elproc);
 *	bool_t	xdr_float(XDR *xdrs, float *fp);
 *	bool_t	xdr_double(XDR *xdrs, double *dp);
 *      bool_t  xdr_string(XDR *xdrs, char **string, u_int maxlen);
 *      bool_t  xdr_bytes(XDR *xdrs, caddr_t *arrp,u_int *sizep, u_int maxsz); 
 *      bool_t  xdr_opaque(XDR *xdrs, char *data, u_int bytes);
 *
 *  ADVANCED
 *
 *      u_int   xdr_getpos(XDR *xdrs);
 *      bool_t  xdr_setpos(XDR *xdrs, u_int posn);
 *
 *  MANAGEMENT
 *
 *      void    xdrstdio_create(XDR *xdrs, FILE *fp, enum xdr_op op);
 *      void    xdrchan_create(XDR *xdrs, Channel *fp, enum xdr_op op);
 *	void	xdrmem_create(XDR *xdrs,void *addr,u_int size,enum xdr_op op);
 *      void    xdr_destroy(XDR *xdrs);
 *
 *  SUPPORT (network <--> host on shorts and longs)
 *
 *	short   htons(short x)
 *	short   ntohs(short x)
 *	short   htonl(long x)
 *	short   ntohl(long x)
 *
 *
 */

#ifndef __Txdr_h
#define __Txdr_h

#include <stdio.h>

#ifndef _ICC
#include <sys/types.h>
#include <rpc/rpc.h>
#else
#include <channel.h>

#define bool_t int
enum xdr_op {XDR_ENCODE=0,XDR_DECODE=1,XDR_FREE=2};
typedef unsigned short u_short;
typedef unsigned int u_int;
typedef unsigned long u_long;

typedef bool_t (*xdrproc_t)();

typedef char *caddr_t;

typedef struct __xdr_s XDR;

struct __xdr_s {
    enum xdr_op	x_op;		/* operation; fast additional param */
    struct xdr_ops {
	bool_t	(*x_getlong)  (XDR*, long *);
	bool_t	(*x_putlong)  (XDR *, long *);
	bool_t	(*x_getbytes) (XDR *, void *, u_int);
	bool_t	(*x_putbytes) (XDR *, void *, u_int);
	u_int	(*x_getpostn) (XDR *);
	bool_t	(*x_setpostn) (XDR *, u_int);
	long *	(*x_inline)   (XDR *, int);
	void	(*x_destroy)  (XDR *);
    } *x_ops;
    caddr_t 	x_public;	/* users' data */
    caddr_t	x_private;	/* pointer to private data */
    caddr_t  	x_base;		/* private used for position info */
    int		x_handy;	/* extra private word */
};

#define xdr_setpos(xdrs,pos) (*(xdrs)->x_ops->x_setpostn)(xdrs,pos)
#define xdr_destroy(xdrs) (*(xdrs)->x_ops->x_destroy)(xdrs)
#endif

#define xdr_getpos(xdrs) (*(xdrs)->x_ops->x_getpostn)(xdrs)

#define EXTERN extern
#define PRIVATE static
#define PUBLIC

#ifdef _ICC
EXTERN short htons(short);
EXTERN long htonl(long);
#define ntohs(s) htons(s)
#define ntohl(l) htonl(l)
#endif

EXTERN bool_t t_xdr_short(XDR *xdrs, short *sp);
EXTERN bool_t t_xdr_int(XDR *xdrs, int *ip);
EXTERN bool_t t_xdr_long(XDR *, long *);
EXTERN bool_t t_xdr_array(XDR *, caddr_t *, u_int *, u_int , u_int, xdrproc_t);
EXTERN bool_t t_xdr_float(XDR *, float *);
EXTERN bool_t t_xdr_double(XDR *, double *);
EXTERN bool_t t_xdr_string(XDR *xdrs, char **, u_int );
EXTERN bool_t t_xdr_bytes(XDR *, caddr_t *, u_int *, u_int );
EXTERN bool_t t_xdr_opaque(XDR *, char *, u_int);
EXTERN void t_xdrmem_create(XDR *, void *, u_int, enum xdr_op);
EXTERN void t_xdrstdio_create(XDR *, FILE *, enum xdr_op);

#define xdr_short t_xdr_short
#define xdr_int t_xdr_int
#define xdr_long t_xdr_long
#define xdr_array t_xdr_array
#define xdr_float t_xdr_float
#define xdr_double t_xdr_double
#define xdr_string t_xdr_string
#define xdr_bytes t_xdr_bytes
#define xdr_opaque t_xdr_opaque
#define xdrmem_create t_xdrmem_create
#define xdrstdio_create t_xdrstdio_create

#ifdef _ICC
EXTERN void t_xdrchan_create(XDR *, Channel *, enum xdr_op);
#define xdrchan_create t_xdrchan_create
#endif

#endif
