/**********************************************************************\
*                                                                      *
*                  Copyright (c) Exeter University,		       *
*                  Department of Computer Science,                     *
*                 (aba@uk.ac.ex.dcs) Adam Back, 1994                   *
*                                                                      *
*  GNU GPL (General Public License) Applies.                           *
*                                                                      *
\**********************************************************************/
/*************************************************************************
*                                                                        *
* Memory management routines providing allocation of arbitrarily         *
* aligned memory blocks.                                                 *
*                                                                        *
*************************************************************************/

/* #define LOG */
#define DEBUG
/* #define STATS */
/* #define CHECK */

#if defined(_ICC) && !defined(TRANSPUTER)
#define TRANSPUTER
#endif

#if defined(TRANSPUTER)
#include <semaphor.h>
#endif

#ifdef DEBUG
#include <stdio.h>
#endif

#include "t_amalloc.h"

#define MIN_BLOCK 16		/* MUST be > sizeof(int) */
				/* malloc(x) st x < MIN_BLOCK is the same as */
				/* malloc(MIN_BLOCK) */
#define GAP 4

typedef struct free_tag free_list_type;

struct free_tag {
  int size;
  free_list_type *next;
};

free_list_type *free_list;
void *arena_start, *arena_end;

#if defined(STATS)
int allocated = 0;
int free_mem = 0;
#endif

#if defined(TRANSPUTER)
Semaphore memLock;
#endif

#if defined(CHECK)
void check(free_list_type* p) {
  int err = 0, i;
  static free_list_type backup[10], t;
  free_list_type* c;
  if ((unsigned int) p % 4 != 0) {
    printf("heap: aligment fault\n");
    err = 1;
  } else if ( ! ( t_in_malloc_arena( p ) ) ) {
    printf("heap: pointer not in heap\n");
    err = 1;
  } else if ( p->size <=0 || p->size > 4*1024*1024 ) {
    printf("heap: garbage size field\n");
    err = 1;
  }
  if ( err ) {
    SemSignal( &memLock );
    c = free_list;
    i = 0;
    while ( i < 10 && c) {
      t = *c;
      *c = backup[ i ];
      backup[ i ] = t;
      c = c->next;
      i++;
    }
    printf("previous----------\n");
    report();
    c = free_list;
    i = 0;
    while ( i < 10 && c) {
      t = *c;
      *c = backup[ i ];
      backup[ i ] = t;
      c = c->next;
      i++;
    }
    printf("current----------\n");
    report();
  } else {
    c = free_list;
    i = 0;
    while ( c && i < 10 ) {
      backup[ i ] = *c;
      c = c->next;
      i++;
    }
  }
}

#define SIZE(x) size_check(x,__LINE__)

void size_check(int size, int line) {
  if ( size <= 0 || size > 4*1024*1024 ) {
    printf("size = %d at line %d\n", size, line);
  }
}

#endif

void t_malloc_init( void *arena , unsigned int arena_size ) {
  free_list = arena;
  free_list->size = 0;

  /* the gap ensures nothing will be joined to the first block */

  free_list->next = (free_list_type*)((char*)arena + sizeof(int) + GAP);
  free_list->next->size = arena_size - 2 * sizeof(int) - GAP;
  free_list->next->next = 0;
  arena_start = arena;
  arena_end = (void*)((char*)arena + arena_size);
#if defined(STATS)
  allocated = 0;
  free_mem = arena_size - 2 * sizeof(int) - GAP;
#endif
#if defined(TRANSPUTER)
  SemInit( &memLock, 1 );
#endif
}

void *t_malloc( unsigned int request ) {
  unsigned int rem;
  free_list_type *prev, *remainder, *ptr;
#if defined(CHECK)
  int n = 0;
/*  if ( request > 16384 ) {
    report();
  }
*/
#endif

#if defined(LOG)
  printf("t_malloc(%d)\n", request);
#endif

  if ( request < MIN_BLOCK ) request = MIN_BLOCK;
  if ( request % 4 != 0 ) request = (request + 3) & ~3;

#if defined(TRANSPUTER)
  SemWait( &memLock );
#endif

  prev = free_list;
  ptr = free_list->next;
  while ( ptr ) {
#if defined(CHECK)
    n++;
    check( ptr );
    if ( n > 10 ) {
      report();
    }
#endif
    if ( request <= ptr->size ) {
      rem = ptr->size - request;
      if ( rem >= MIN_BLOCK + sizeof(int) ) { /* not exact match */

/*  remainder >= MIN_BLOCK
 *
 *  +----+---------+----+-----------+
 *  | sz | request | sz | remainder |
 *  +----+---------+----+-----------+
 *
 */
	remainder = (free_list_type*)((char*)ptr + request + sizeof(int));
	remainder->size = rem - sizeof(int);
#if defined(CHECK)
	SIZE(remainder->size);
#endif
	remainder->next = ptr->next;
	prev->next = remainder;
	ptr->size = request;
#if defined(CHECK)
	SIZE(ptr->size);
#endif
#if defined(STATS)
	free_mem -= request + sizeof(int);
	allocated += request + sizeof(int);
#endif
      } else {
/*  remainder < MIN_BLOCK
 *
 *  +----+-----------------------+
 *  | sz | request  +  remainder |
 *  +----+-----------------------+
 *
 */
	prev->next = ptr->next;
#if defined(STATS)
	free_mem -= ptr->size + sizeof(int);
	allocated += ptr->size + sizeof(int);
#endif
      }
#if defined(TRANSPUTER)
      SemSignal( &memLock );
#endif
      return (void*)((char*)ptr + sizeof(int));
    }
    prev = ptr;
    ptr = ptr->next;
  }
#if defined(TRANSPUTER)
  SemSignal( &memLock );
#endif
  return 0;
}

void *t_amalloc( unsigned int request, unsigned int alignment ) {
  unsigned int dist, rem;
  free_list_type *prev, *block, *remainder, *ptr;

#if defined(LOG)
  printf("t_amalloc(%d, %d)\n", request, alignment);
#endif

  if ( request < MIN_BLOCK ) request = MIN_BLOCK;
  if ( alignment < sizeof(int) ) alignment = sizeof(int);
  if ( request % 4 != 0 ) request = (request + 3) & ~3;

#if defined(TRANSPUTER)
  SemWait( &memLock );
#endif

  prev = free_list;
  ptr = free_list->next;
  while ( ptr ) {
#if defined(CHECK)
    check( ptr );
#endif
    dist = alignment - ( (unsigned int)ptr + sizeof(int) ) % alignment;
    if ( dist == 0 ) {		/* already aligned */

      if ( request <= ptr->size ) {
	rem = ptr->size - request;
	if ( rem >= MIN_BLOCK + sizeof(int) ) {

/*  aligned && remainder >= MIN_BLOCK
 *
 *    aligned
 *       |
 *  +----+---------+----+-----------+
 *  | sz | request | sz | remainder |
 *  +----+---------+----+-----------+
 *       |
 */
	  remainder = (free_list_type*)((char*)ptr + request + sizeof(int));
	  remainder->size = rem - sizeof(int);
#if defined(CHECK)
	  SIZE(remainder->size);
#endif
	  prev->next = remainder;
	  remainder->next = ptr->next;
	  ptr->size = request;
#if defined(CHECK)
	  SIZE(remainder->size);
#endif

#if defined(STATS)
	  free_mem -= rem + sizeof(int);
	  allocated -= rem + sizeof(int);
#endif
	} else {

/*  aligned && remainder < MIN_BLOCK
 *
 *    aligned
 *       |
 *  +----+-----------------------+
 *  | sz | request  +  remainder |
 *  +----+-----------------------+
 *       |
 */
	  prev->next = ptr->next;
#if defined(STATS)
	  free_mem -= ptr->size + sizeof(int);
	  allocated += ptr->size + sizeof(int);
#endif
	}
#if defined(TRANSPUTER)
	SemSignal( &memLock );
#endif
	return (void*)((char*)ptr + sizeof(int));
      }
    } else {			/* not aligned */

      dist = alignment - 
	((unsigned int)ptr + 2 * sizeof(int) + MIN_BLOCK ) % alignment;
      if ( dist + request + sizeof(int) + MIN_BLOCK <= ptr->size ) {
	rem = ptr->size - dist - request - sizeof(int) - MIN_BLOCK;

	ptr->size = dist + MIN_BLOCK;
#if defined(CHECK)
	SIZE(ptr->size);
#endif
	block = (free_list_type*)((char*)ptr + dist + MIN_BLOCK + sizeof(int));
	if ( rem > MIN_BLOCK + sizeof(int) ) {

/*  !aligned && remainder >= MIN_BLOCK
 *
 *               aligned
 *                  |
 *  +----+-----+----+---------+----+-----------+
 *  | sz | pad | sz | request | sz | remainder |
 *  +----+-----+----+---------+----+-----------+
 *                  |
 *
 *  note: pad is defined to be >= MIN_BLOCK
 *
 */
	  remainder=(free_list_type*)((char*) ptr + dist + request +
				      MIN_BLOCK + 2 * sizeof(int));
	  remainder->size = rem - sizeof(int);
#if defined(CHECK)
	  SIZE(remainder->size);
#endif
	  remainder->next = ptr->next;
	  ptr->next = remainder;
	  block->size = request;
#if defined(CHECK)
	  SIZE(remainder->size);
#endif
	} else {

/*  !aligned && remainder < MIN_BLOCK
 *
 *               aligned
 *                  |
 *  +----+-----+----+-----------------------+
 *  | sz | pad | sz | request  +  remainder |
 *  +----+-----+----+-----------------------+
 *                  |
 *
 */
	  block->size = request + rem;
#if defined(CHECK)
	  SIZE(block->size);
#endif
	}
#if defined(TRANSPUTER)
	SemSignal( &memLock );
#endif
	return (void*)((char*)block + sizeof(int));
      }
    }
    prev = ptr;
    ptr = ptr->next;
  }
#if defined(TRANSPUTER)
  SemSignal( &memLock );
#endif
  return 0;
}

void t_free( void* pointer ) {
  int joined = 0;

  free_list_type *ptr, *block, *prev;

#if defined(LOG)
  printf("t_free(0x%08x)\n", pointer);
#endif

  if ( pointer == 0 ) return;

  block = (void*)((char*)pointer - sizeof(int));

#if defined(TRANSPUTER)
  SemWait( &memLock );
#endif

  prev = free_list;
  ptr = free_list->next;
#if defined(CHECK)
  check( ptr );
#endif
  while ( ptr && (unsigned int)ptr < (unsigned int)block ) {
    prev = ptr;
    ptr = ptr->next;
#if defined(CHECK)
  check( ptr );
#endif
  }
				/* join with prev block */
  if ( (char*)prev + prev->size + sizeof(int) == (char*)block ) { 
    prev->size += block->size + sizeof(int);
    joined = 1;
  }

  if ( !ptr ) {
#if defined(TRANSPUTER)    
    SemSignal( &memLock );
#endif
    return;
  }
				/* join with next block */
  if ( (char*)block + block->size + sizeof(int) == (char*)ptr ) {
    if ( joined ) {
      prev->size += ptr->size + sizeof(int); /* join both */
      prev->next = ptr->next;
    } else {
      block->size += ptr->size + sizeof(int);
      block->next = ptr->next;
      prev->next = block;
      joined = 1;
    }
  }
  if ( !joined ) {
    block->next = ptr;
    prev->next = block;
  }
#if defined(TRANSPUTER)
  SemSignal( &memLock );
#endif
}

/* MACRO IMPLEMENTATION of t_in_malloc_arena */
/*
int t_in_malloc_arena( void *ptr ) {
  if ((char*)ptr >= (char*)arena_start && (char*)ptr <= (char*)arena_end ) {
    return 1;
  } else {
    return 0;
  }
}
*/

int t_arena_size(void) {
  return (char*)arena_end - (char*)arena_start;
}

#if defined(DEBUG)
void report( void ){
  free_list_type *ptr;
  
#if defined(TRANSPUTER)
  SemWait( &memLock );
#endif
  ptr = free_list->next;
  while ( ptr ) {
    fprintf( stderr, "\t    +-----------+\n" );
    fprintf( stderr, "0x%08x  |%d\t|\n", ptr, ptr->size );
    ptr = ptr->next;
  }
  fprintf( stderr, "\t    +-----------+\n" );

#if defined(TRANSPUTER)
  SemSignal( &memLock );
#endif
}
#endif
